# XfwfEntry - single line text input widget
# Original Author: 
#      AF = Andre Fornacon (afo@pub.th-zwickau.de)
# Modifications by:
#      PK = Paul Kendall <paul@kcbbs.gen.nz> 
#      LW = Lachlan Wetherall <lwether@cs.newcastle.edu.au>
#
# Revision history:
# Ver   Date          Comments
# ---   ----          --------
# 1.0      Jan-1995   Initial release (AF)
# 1.1      Feb-1995   Now child of Board (I need too less from Label) (AF)
# 1.2   05-Jul-1995   Modified to handle variable width fonts. (PK)
# 1.3   01-Aug-1995   Added modifyCallback,XfwfEntryCallbackStruct and modified
#                     self_insert() to call the modifyCallback before insertion
#                     is done (AF)
# 1.4   10-Sep-1996   Made insert cursor a vertical bar.
#                     Added blinkingCursor and blinkRate. 
#                     Added scrolling for text longer than the field size 
#                     Added text selection/pasting 
#                     Fixed bugs in del_char, backspace and insert_char. (LW)
# 1.5   16-Sep-1996   Added 'afterModifyCallback' (LW)
#
#
# TODO: 
#   - extend selection action
#   - scroll field when dragging a selection out
#   - make double clicking select by word


@class XfwfEntry (XfwfBoard)	@file=Entry	

@ The |XfwfEntry| widget is used to read a single line of text.
You can restrict the total length of the input string using the 
variable |max| . The default value is 30 characters.
Using the |valid| resource it is possible to control the users input.
The user can only enter the specified characters.
Setting this ressource to NULL (default) let the users input all
characters between 0x20 (SPACE) and 0xff.
The |echo| resource enables/disables displaying of the entered text.
You can use this for things like password boxes.


@PUBLIC


@ The variable |text| holds the string modified by the user.

	@var String text = NULL
	
	
@ The |activate| callback is called if the activate action is executed,
bound by default to the <Return> key. Entered text is passed in the 
|call_data| argument.

	@var <Callback> XtCallbackList activate = NULL


@ The |modifyCallback| handler is called before the insert is done.
The handler is called with a XfwfEntryCallbackStruct as call_data.

    @var <Callback> XtCallbackList modifyCallback = NULL


@ The |afterModifyCallback| handler is called after any change (insertion
or deletion) is made to the text.
The handler is called with a XfwfEntryCallbackStruct as call_data.

    @var <Callback> XtCallbackList afterModifyCallback = NULL


@ Using the |valid| resource it is possible to limit the characters the user 
is allowed to enter. With a value of NULL the allowed characters ranges
from 0x20 (SPACE) to 0xff.

	@var String valid = NULL


@ The |max| resource defines the maximum number of characters the widget can
accept. It defaults to 30.

	@var int max	= 30


@ The |echo| flag enables/disables displaying of the entered text.

	@var Boolean echo = True


@ The text is drawn using the font given in the |font| resource. The font can
be monospaced or proportional.

	@var <FontStruct> XFontStruct *font = <String> XtDefaultFont

@ The |foreground| resource specifies the color used for drawing the text.

	@var Pixel foreground = <String> XtDefaultForeground


@ If |True|, the insert cursor will blink. The default value is |False|.

	@var Boolean blinkingCursor = False

@ The insert cursor blinking interval in milliseconds. The default value
is 800ms.

	@var int blinkInterval = 800

@ The |selectThreshold| resource specifies how many pixels the mouse cursor
must be moved before text selection is activated.

	@var int selectThreshold = 3

@ If the |killSelection| resource is true, then typing a character when a
selection is highlighted will cause the selection to be deleted first. If the
resource is false, the selection is simply unhighlighted. The default value
is true.

	@var Boolean killSelection = True

@ This widget receives keyboard events, |traversalOn| is set to |True|.

	@var traversalOn = True

@ The default frame width is set to 2.

	@var frameWidth = 2

@ The |frameType| resource defaults to |"sunken"|.

	@var frameType = XfwfSunken


@PRIVATE

@ The current length of the text

	@var int text_len
	
@ current cursor position	

	@var int text_pos

@ The |text_start| ressource indicates the first displayed character , 
if the whole exceeds the width . Needed for horizontal scrolling.

	@var int text_start

@ Between keypresses, the |XComposeStatus| is retained.

	@var XComposeStatus compose_stat

@ For normal text, ngc is used

	@var GC ngc

@ For drawing text in reverse color (like the cursor) rgc is used

	@var GC rgc

@ The graphics context for drawing the insert cursor.

	@var GC cursorGC

@ Remember whether the insert cursor is currently active.

	@var Boolean cursorActive

@ Remember whether the insert cursor is currently visible.

	@var Boolean cursorVisible

@ A timer form implementing cursor blinking

	@var XtIntervalId blinktimer

@ The variable |selecting| is True if the user is currently selecting text with
mouse button 1.

	@var Boolean selecting


@ The variable |selected| is True if this widget holds the current text 
selection.

	@var Boolean selected

@ The mouse position in the widget when text selection was started.

	@var int selectx
	@var int selecty


@ The starting and ending positions in the current text selection.

	@var int selectStart
	@var int selectEnd

@EXPORTS

@ If |modifyCallback| or |afterModifyCallback| is set, the callback handler is 
called with a pointer to a XfwfEntryCallbackStruct as call_data. 

The |modifyCallback| function is called before the insertion takes place.
The insertion is aborted if the flag |doit| is set to False (default:True)
or text is NULL. The field |text| is the string to be inserted and |length| is 
the number of chars in the string.
|text| points to a memory area which is malloc'd before |modifyCallback|
is called and is free'd after the call - that means,if you change the text
in the callback handler you have to free the original passed text and
provide the new text as an malloc'd area.
In the case of setting text to NULL you have only to free it.

The |afterModifyCallback| is called after every change (insertion or deletion)
to the widget's text. The field |text| points to a null terminated string 
of the widgets current text, and |length| contains the length of the string.
The |text| should not be free'd or manipulated by the callback function.
The field |doit| is ignored for the |afterModifyCallback|.

       @type XfwfEntryCallbackStruct = struct {
                       Boolean doit;
                       String text;
                       int length;
       }

@ Clear the text entry from memory and the display.

@proc void XfwfEntryClear($)
{
	Position x,y;
	Dimension h,w;
	
	if($text != NULL) {
		get_pos($,&x,&y,&w);
		h=$font->ascent+$font->descent+3; /* +3 in order to clear insertion cursor */
#ifdef DEBUG		
		fprintf(stderr,"clear(%s,%d) %dx%d @%d,%d,[%dx%d]\n",$text,$text_len,\
				w,h,x,y,$width,$height);
#endif		
		$text_pos=$text_len=$text_start=0;
		$text[0]='\0';
		$selected=False;
		if (w>0) {
			XClearArea(XtDisplay($),XtWindow($),x-2,y-$font->ascent-1,w+4,h,False);
			$cursorVisible=False;
			}
		if($traversal_focus == True)
			set_cursor($,True);
	}
}

@METHODS

@ The |initialize| method sets the private variables to default values

@proc initialize
{
	String tmp;
	int i;

#ifdef DEBUG	
	fprintf(stderr,"init(%s) [%s,%d,%s]\n",XtName($),$valid,$max,\
			$echo ? "true" : "false");
#endif
	
	$text_pos=$text_len=$text_start=0;

	$ngc=$rgc=$cursorGC=NULL;
	make_gc($);
	$cursorActive=$cursorVisible=False;
	$blinktimer=0;

	/* nothing is selected initially */
	$selecting=False;
	$selected=False;
	$selectStart=$selectEnd=0;
	
	if($valid != NULL)
		$valid=XtNewString($valid);
	tmp=XtMalloc((1+$max)*sizeof(char));
	tmp[0]='\0';

	if($text != NULL)
	{
		for(i=0;$text[i] && i < $max;i++)
			tmp[i]=$text[i];
		tmp[i]='\0';
		$text_pos=$text_len=i;
	}
	$text=tmp;
#ifdef DEBUG	
	fprintf(stderr,"entry.init(%s)\n",$text);
#endif	
}


@ The |set_values| method copies all changed string ressources 

@proc set_values
{
	String tmp;
	int i;

#ifdef DEBUG	
	fprintf(stderr,"set_values(%s) '%s'\n",XtName($),$text);
#endif
	
	if($old$text != $text)
	{
		XtFree($old$text);
		tmp=XtMalloc((1+$max)*sizeof(char));
		for(i=0;$text[i] && i < $max;i++)
			tmp[i]=$text[i];
		tmp[i]='\0';
		$text_pos=$text_len=i;
		$text=tmp;
	}
	if ($old$blinkingCursor != $blinkingCursor) 
		set_cursor($, $cursorActive);
		

	return True;	
}

@ The |expose| method is called if the widget was overlapped and became
	visible again
	
@proc expose
{
	if(! XtIsRealized($))
		return ;
	
	#expose($,event,region);
	set_text($,0);
	if($traversal_focus == True)
		set_cursor($,True);
	
#ifdef DEBUG	
	fprintf(stderr,"expose(%s)\t,%s,len %d, max %d, pos %d\n",XtName($),\
				$text,$text_len,$max,$text_pos);
#endif	
}


@TRANSLATIONS

@trans <FocusIn>:			focusIn() cursor_on()
@trans <FocusOut>:			cursor_off() focusOut()
@trans <EnterWindow>:		enter_window() 
@trans <LeaveWindow>:		leave_window()

@trans <Btn1Down>:			traverseCurrent() mclick() start_selection()
@trans <Btn1Up>:            end_selection() 
@trans <Btn1Motion>:        move_selection()
@trans <Btn2Down>:			traverseCurrent() pasteSelection()
@trans <Key>Up:				traverseUp()
@trans <Key>Down:			traverseDown()

@trans	Shift<Key>Tab:		traversePrev()
@trans	~Shift<Key>Tab:		traverseNext()

@trans	<Key>Delete:		del_selection()
@trans	<Key>BackSpace:		backspace()
@trans	<Key>Right:			right()
@trans	<Key>Left:			left()

@trans	<Key>Home:			bol()
@trans	<Key>End:			eol()

@trans	Ctrl<Key>a:			bol()
@trans	Ctrl<Key>d:			del_char()
@trans	Ctrl<Key>e:			eol()
@trans	Ctrl<Key>b:			left()
@trans	Ctrl<Key>f:			right()
@trans	Ctrl<Key>h:			backspace()
@trans	Ctrl<Key>k:			kill()
@trans	Ctrl<Key>l:			redraw()
@trans	<Key>Return:		activate()
@trans	~Ctrl<Key>:			self_insert()


@ACTIONS

@ The |activate| action calls the |activate| callback with the current
|text| as |call_data|.

@proc activate {
    XtCallCallbackList($, $activate, $text);
	}

@ The |self_insert| action inserts the passed character into the entry field
at the current insertion point as long as the maximum field length has not
been reached. If a |modifyCallback| has been registered, it is called before
the insertion takes place.
If a |afterModifyCallback| has been registered, it is called after
the insertion takes place.

@proc self_insert {
    KeySym keysym;
    int n,k,i;
    char *buf;
    XfwfEntryCallbackStruct cbs;

    if (event->type != KeyPress && event->type != KeyRelease) {
		XtWarning("action <self_insert> can only handle keyboard events");
		return;
	    }


    buf=XtMalloc((1+$max)*sizeof(char));
    n = XLookupString(&event->xkey, buf, $max, &keysym, &$compose_stat);
    if($modifyCallback) {
         cbs.doit=True;
         cbs.text=buf;
         cbs.length=n;
         XtCallCallbackList($, $modifyCallback, &cbs);

         if(cbs.doit == False) {
              XtFree(buf);
              /* printf("\tabort modify (doit==False)\n"); */
              return;
              }

         if(cbs.text == NULL)
              return;
         buf=cbs.text;
         n=cbs.length;
         }

	set_cursor($,False);
	for(k=i=0;i < n;i++)
		k += insert_char($,buf[i]);

	if (k>0)
		set_text($,$text_pos-k);
	set_cursor($,True);
    XtFree(buf);

    if($afterModifyCallback) {
		cbs.text=$text;
		cbs.length=$text_len;
		XtCallCallbackList($, $afterModifyCallback, &cbs);
		}

#ifdef DEBUG	
	fprintf(stderr,"self_insert()\t%d ->%s<\n",keysym,$text);
#endif	
	}


@ Delete the character to the right of the cursor.

@proc del_char {
    int i;
	XfwfEntryCallbackStruct cbs;

	clear_selection($);
	if($text_pos < $text_len) {
		set_cursor($,False);
		$text_len--;
		for(i=$text_pos;i < $text_len; i++)
			$text[i]=$text[i+1];
		$text[$text_len]='\0';
		if($text_pos > $text_len)
			$text_pos=$text_len;
		set_text($,$text_pos);
		set_cursor($,True);

	    if($afterModifyCallback) {
			cbs.text=$text;
			cbs.length=$text_len;
			XtCallCallbackList($, $afterModifyCallback, &cbs);
			}
		}	
	}

@ Delete the current text selection if it exists - delete the character to the 
right of the cursor otherwise.

@proc del_selection {
    int i;
    XfwfEntryCallbackStruct cbs;
	Boolean changed=False;

	/* if there is a selection, delete it - otherwise delete char to
	   the right of the cursor */

	if ($selected) {
		delete_selection($);
		set_text($,$text_start);
		set_cursor($,True);
		changed=True;
		}
	else if($text_pos < $text_len) {
		set_cursor($,False);
		$text_len--;
		for(i=$text_pos;i < $text_len; i++)
			$text[i]=$text[i+1];
		$text[$text_len]='\0';
		if($text_pos > $text_len)
			$text_pos=$text_len;
		set_text($,$text_pos);
		set_cursor($,True);
		changed=True;
		}	

    if($afterModifyCallback && changed) {
		cbs.text=$text;
		cbs.length=$text_len;
		XtCallCallbackList($, $afterModifyCallback, &cbs);
		}
	}


@ Delete the character to the left of the cursor.

@proc backspace {
    int i;
    XfwfEntryCallbackStruct cbs;

	clear_selection($);
    if($text_len > 0 && $text_pos > 0) {
		set_cursor($,False);
		for(i=$text_pos-1;i < $text_len; i++)
			$text[i]=$text[i+1];
		$text_len--;
		$text_pos--;
		set_text($,$text_pos);
		set_cursor($,True);

	    if($afterModifyCallback) {
			cbs.text=$text;
			cbs.length=$text_len;
			XtCallCallbackList($, $afterModifyCallback, &cbs);
			}
		}
	}


@ Delete from current cursor position to end of line.

@proc kill {
	Position x,y;
	Dimension w,h;
    XfwfEntryCallbackStruct cbs;
	
	clear_selection($);
	if($text_pos < $text_len) {
		get_pos($,&x,&y, NULL);
		x+=XTextWidth($font,$text,$text_pos);
		w=XTextWidth($font,$text+$text_pos,(1+$text_len-$text_pos));
		h=$font->ascent+$font->descent;
		
#ifdef DBEUG		
		fprintf(stderr,"kill(%s,%d,%d) %dx%d @%d,%d\n",$text,$text_pos,\
				$text_len,w,h,x,y);
#endif
		
		XClearArea(XtDisplay($),XtWindow($),x,y-$font->ascent,w,h,False);
		$text[$text_pos]='\0';
		$text_len=$text_pos;
		set_text($,$text_pos);
		set_cursor($,True);

	    if($afterModifyCallback) {
			cbs.text=$text;
			cbs.length=$text_len;
			XtCallCallbackList($, $afterModifyCallback, &cbs);
			}
		}
	}


@ Redraw all the text visible in the entry field.

@proc redraw {
	set_text($,0);
	set_cursor($,True);
	}


@ Move the insert cursor to the position of the mouse click.

@proc mclick {
	Position x,y;
	XButtonEvent xbutton;
	
	if(event->type != ButtonPress) {
		fprintf(stderr,"mclick called with invalid event type\n");
		XBell(XtDisplay($),100);
		return;
		}

	xbutton= event->xbutton;
	set_cursor($,False);
	get_pos($,&x,&y,NULL);

	for($text_pos=$text_start; $text_pos<=$text_len 
		&& XTextWidth($font,$text+$text_start,$text_pos-$text_start)<(xbutton.x-x); 
		$text_pos++);
	$text_pos--;
	if($text_pos > $text_len)
		$text_pos=$text_len;
	set_cursor($,True);

#ifdef DEBUG
	fprintf(stderr,"mclick @%d,%d -> %d\n",xbutton.x,xbutton.y,$text_pos);
#endif	
	}

@ Turn the insertion cursor on.

@proc cursor_on {
	set_cursor($,True);	
	}

@ Turn the insertion cursor off.

@proc cursor_off {
	set_cursor($,False);	
	}


@ If keyboard traversal is off, this action routine will highlight the
border and turn the insertion cursor on when the mouse pointer enters the 
widget.

@proc enter_window {
	if ($traversalOn)
		return;
	set_cursor($,True);	
	$highlight_border($);
	}


@ If keyboard traversal is off, this action routine will unhighlight the
border and turn the insertion cursor off when the mouse pointer leaves the 
widget.

@proc leave_window {
	if ($traversalOn)
		return;
	set_cursor($,False);	
	$unhighlight_border($);
	}


@ Move to the beginning of the line.
	
@proc bol {
	if ($text_pos != 0)	{
		set_cursor($,False);		
		$text_pos = 0;
		set_cursor($,True);		
		}
	}


@ Move to the end of the line.

@proc eol {
	if ($text_pos < $text_len) {
		set_cursor($,False);
		$text_pos = $text_len;
		set_cursor($,True);
		}
	}


@ Move left one character. 

@proc left {
	if($text_pos > 0) {
		set_cursor($,False);
		$text_pos--;
		set_cursor($,True);
		}
	}


@ Move right one character. 

@proc right {
	if($text_pos < $text_len) {
		set_cursor($,False);
		$text_pos++;
		set_cursor($,True);
		}
	}


@ Paste the current X11 text selection at the insertion point.

@proc pasteSelection {
	XtGetSelectionValue($, XA_PRIMARY, XA_STRING, paste_callback, NULL,
		event->xbutton.time);
	}


@ Start text selection. (Usually bound to mouse button 1.)

@proc start_selection {

	/* don't start selecting until the mouse has moved a certain number
	   of pixels */
	$selecting=False;
	clear_selection($);


	/* remember where the mouse is */
    $selectx=event->xbutton.x;

	find_position($, $selectx, &($selectStart));
	$selectEnd=$selectStart;
	}

@ Drag out the text selection. (Usually bound to motion while mouse button 1
is held down.

@proc move_selection {
	int mx, lastend;
	Boolean selection_started=False;

	mx=event->xbutton.x;
	if (! $selecting)
		/* has the mouse moved far enough to start selection? */
		if (abs(mx-$selectx) >= $selectThreshold) {
			$selected=True;
			$selecting=True;
			selection_started=True;
			set_cursor($, False);
			}

	if (! $selecting)
		return;

	lastend=$selectEnd;
	find_position($, mx, &($selectEnd));
	/* if the selection end has moved (or just begun) 
	   we need to update the display */
	if (($selectEnd != lastend) || selection_started) {
		draw_text($, lastend, $selectEnd);
		}
	}


@ Finish making a text selection with the mouse.

@proc end_selection {
	if (!$selecting)
		return;

	$selecting=False;
	$text_pos=$selectEnd+1;
	set_cursor($, True);

	if (XtOwnSelection($, XA_PRIMARY, event->xbutton.time, 
		convert_proc, lose_ownership_proc, NULL) == False) {
		XtWarning("Entry: failed to become selection owner\n");
		return;
		}
	$selected=True;
	}

@UTILITIES

@ Called if another widget obtains the ownership of the PRIMARY atom.

@proc lose_ownership_proc($, Atom *selection) {
	clear_selection($);
	}


@ The procedure to convert the current text selection.

@proc Boolean convert_proc($, Atom *selection, Atom *target, Atom *type, 
        XtPointer *value, unsigned long *length, int *format) {
	int s1, s2;
#ifdef SVR4                                     /* BB, 27 Mar '96 */
#define bcopy(b1, b2, length) memcpy(b2, b1, length)
#endif
    Display* d = XtDisplay($);
    XSelectionRequestEvent* req = XtGetSelectionRequest($, *selection, NULL);

    if (*target == XA_TARGETS(d)) {
        Atom* targetP;
        Atom* std_targets;
        unsigned long std_length;
        XmuConvertStandardSelection($, req->time, selection, target,
                                    type, (XPointer*)&std_targets,
                                    &std_length, format);
        *value = XtMalloc(sizeof(Atom) * (std_length + 1));
        targetP = *(Atom**) value;
        *targetP++ = XA_STRING;
        *length = std_length + (targetP - (*(Atom **) value));
        bcopy((char*)std_targets, (char*)targetP, sizeof(Atom)*std_length);
        XtFree((char*)std_targets);
        *type = XA_ATOM;
        *format = 32;
        return True;
    }
    if (*target == XA_STRING) {
        *type = XA_STRING;
		if (! $selected)
			return False;
		s1=min($selectStart, $selectEnd);
		s2=max($selectStart, $selectEnd);
        *value = XtMalloc(s2-s1+2);
		strncpy(*value, $text+s1, s2-s1+1); 
        *length = sizeof(char) * (s2-s1+1);
        *format = 8;
        return True;
    }
    if (XmuConvertStandardSelection($, req->time, selection, target,
                                    type, (XPointer*) value, length, format))
        return True;

    return False;
}

@ Clears the text selection if it was active.

@proc clear_selection($) {
	if (!$selected) 
		return;

	$selected=False;

	/* redisplay the text */
	set_text($, $text_start); 
	XtDisownSelection($, XA_PRIMARY, CurrentTime);
	}


@ Deletes the text selection if it was active.

@proc delete_selection($) {
	int s1, s2;
	char *dst, *src;

	if (!$selected) 
		return;
	$selected=False;

	s1=min($selectStart, $selectEnd);
	s2=max($selectStart, $selectEnd);
	dst=$text+s1;
	src=$text+s2+1;
	while (*src != '\0') 
		*dst++ = *src++;
	*dst='\0';
	$text_pos=s1;
	$text_len=strlen($text);
	set_text($, $text_start);
	XtDisownSelection($, XA_PRIMARY, CurrentTime);
	}


@ Given a mouse position |(x,y)| this procedure computes which 
character position that the mouse position corresponds to.

@proc find_position($, int x, int *marker) {
	int i, xpos, slen;
	Position wx, wy;
	
	slen=strlen($text+$text_start);
	get_pos($, &wx, &wy, NULL);

	for (i=1; i<slen; i++) {
		xpos=wx + XTextWidth($font,$text+$text_start,i);
		if (xpos>=x)
			break;
		}

	*marker=i-1+$text_start;
	}


@ The |paste_callback| is installed by the |pasteSelection|
action procedure. It receives the selection data and `pastes' it
into the Entry widget. In this case that means it is passed
character for character to the |insert_char| function.

@proc paste_callback($, XtPointer client_data, Atom *selection, Atom *type, 
	XtPointer value, unsigned long *length, int *format) {
    char *text = (char *) value;
	int old_text_pos, i;

	assert(*selection == XA_PRIMARY);

	if (value == NULL && *length == 0) {
        XBell(XtDisplay($), 0);
        return;
    }

	set_cursor($, False);
	old_text_pos=$text_pos;
    for (i = 0; i < *length; i++) 
        insert_char($, text[i]);
		
    XtFree(value);
	set_text($,old_text_pos);
	set_cursor($, True);
	}


@ create the GC's used for drawing text

@proc make_gc($) {
	XtGCMask mask;
	XGCValues values;

	if($ngc != NULL)				/* GC for normal text */
		XtReleaseGC($,$ngc);

	values.foreground=$foreground;
	values.background=$background_pixel;
	values.font=$font->fid;
	mask=GCFont|GCBackground|GCForeground;
	$ngc=XtGetGC($,mask,&values);

	if($rgc != NULL)				/* GC for reverse text */
		XtReleaseGC($,$rgc);

	values.foreground=$background_pixel;	
	values.background=$foreground;			
	values.font=$font->fid;
	mask=GCFont|GCBackground|GCForeground;
	$rgc=XtGetGC($,mask,&values);

	if($cursorGC != NULL)				/* GC for insert cursor */
		XtReleaseGC($,$cursorGC);

	values.foreground=$foreground ^ $background_pixel;	
	values.background=$background_pixel;			
	values.function=GXxor;
	mask=GCFunction|GCBackground|GCForeground;
	$cursorGC=XtGetGC($,mask,&values);
	}	


@ Get the Position of the first char and the width of the drawable area.
If the width is nott required, a NULL can be passed instead.

@proc get_pos($,Position *x, Position *y, Dimension *w) {
	Dimension baseline;
	XRectangle rect;
	
	baseline = $font->ascent + $font->descent;
	$compute_inside($, &rect.x, &rect.y, &rect.width, &rect.height);

	rect.x += 2;	rect.width -= 4;
	rect.y += 2;	rect.height -= 4;

	*y = rect.y + (rect.height - baseline)/2 + $font->ascent;
	*x = rect.x;
	if (w != NULL)
		*w = rect.width;
}



@proc set_cursor($, Boolean cursoron) {
	$cursorActive = cursoron;
	if ($blinktimer)
		XtRemoveTimeOut($blinktimer);

	if (adjust_start($)) {
		set_text($, 0);  /* redraw all the text */
		$cursorVisible=False;
		}

	if ($cursorActive != $cursorVisible)
		invert_cursor($);

	if ($blinkingCursor && $cursorActive)
		 $blinktimer = XtAppAddTimeOut(XtWidgetToApplicationContext($), 
				$blinkInterval/2, flash_cursor, $);
}

@proc flash_cursor(XtPointer client_data, XtIntervalId *timer) {
    Widget $ = (Widget) client_data;

	invert_cursor($);
	$blinktimer = XtAppAddTimeOut(XtWidgetToApplicationContext($), 
		$blinkInterval/2, flash_cursor, $);
	}

@proc invert_cursor($) {
	Position x,y;
	Dimension pos,w;

	/* cursor position is always 0 if echo disabled */
	pos = $echo ? $text_pos-$text_start : 0;
	
	get_pos($,&x, &y, NULL);
	w=XTextWidth($font, $text+$text_start, pos);

	XDrawLine(XtDisplay($), XtWindow($), $cursorGC, x+w, 
		y+$font->descent+1, x+w, y-$font->ascent-1);
	XDrawLine(XtDisplay($), XtWindow($), $cursorGC, x+w-2, 
		y+$font->descent+1, x+w+2, y+$font->descent+1);
	XDrawLine(XtDisplay($), XtWindow($), $cursorGC, x+w-2, 
		y-$font->ascent-1, x+w+2, y-$font->ascent-1);
	$cursorVisible= ! $cursorVisible;
	}


@ Computes where the cursor will be displayed, and if necessary adjusts the
value of |text_start| so that the cursor will be visible. 
Returns true if |text_start| was modified.

@proc Boolean adjust_start($) {
	Position x,y, rightedge, cursor;
	Dimension w;
	int old_text_start=$text_start;

	if ($text_pos<$text_start) {
		$text_start=$text_pos;
		return True;
		}

	get_pos($,&x, &y, &w);
	rightedge = x+w;
	cursor = x+XTextWidth($font,$text+$text_start,$text_pos-$text_start);
	while (cursor>rightedge) {
		$text_start += 1;
		cursor = x+XTextWidth($font,$text+$text_start,$text_pos-$text_start);
		}

	return old_text_start != $text_start;
	}

@proc set_text($,int pos) {
	Position clearx, x,y, rightedge;
	Dimension w, w1, h;
	int numchars;
	
	if($echo == True) {
		if (adjust_start($))
			pos=$text_start;
		if (pos<$text_start)
			pos=$text_start;
		get_pos($,&x, &y, &w);
		rightedge = x+w;
		if (pos<=$text_start) {
			clearx=x-2;
			w += 2;
			}
		else
			clearx =x+XTextWidth($font,$text+$text_start,pos-$text_start);
		h=$font->ascent+$font->descent+3;
		w=rightedge-clearx;
		
		XClearArea(XtDisplay($),XtWindow($), clearx, y-$font->ascent-1,w+2,h,False);
		$cursorVisible=False;

		if (pos>$text_start)
			x += XTextWidth($font, $text+$text_start, pos-$text_start);
	
#ifdef DEBUG 	
		fprintf(stderr,"set_text(%s,%d,%d) @%d,%d\n",$text, pos, $text_start,
			x,y);
#endif

		numchars=$text_len-pos;
		w1=XTextWidth($font, $text+pos, numchars);
		while ((x+w1 > rightedge) && (numchars>0)) {
			numchars--;
			w1=XTextWidth($font, $text+pos, numchars);
			}
		if (numchars>0)
			draw_text($, pos, pos+numchars-1);
	}
}

@ Draws the text from position |p1| to position |p2| inclusive, taking care to
draw selected text in reverse colors.

@proc draw_text($, int p1, int p2) {
	int s1, s2, r1, r2, tmp;
	Position x, y;
	Position p1_x, r1_x, s2_x;

	if (p1>p2) {
		tmp=p1;
		p1=p2;
		p2=tmp;
		}

	get_pos($, &x, &y, NULL);
	p1_x=x;
	if (p1>$text_start)
		p1_x = x + XTextWidth($font, $text+$text_start, p1-$text_start);

	if (! $selected) {
		XDrawImageString(XtDisplay($),XtWindow($),$ngc,p1_x,y,$text+p1,p2-p1+1);
		return;
		}
	s1=min($selectStart, $selectEnd);
	s2=max($selectStart, $selectEnd);

	if (p1<s1)
		XDrawImageString(XtDisplay($),XtWindow($),$ngc,p1_x, y, $text+p1, s1-p1);
	r1=max(p1,s1);
	r2=min(p2,s2);
	if (r1<=r2) {
		r1_x = x + XTextWidth($font, $text+$text_start, r1-$text_start);
		XDrawImageString(XtDisplay($),XtWindow($), $rgc, r1_x, y, $text+r1, r2-r1+1);
		}
	if (p2>s2) {
		s2_x = x + XTextWidth($font, $text+$text_start, s2+1-$text_start);
		XDrawImageString(XtDisplay($),XtWindow($),$ngc, s2_x, y, $text+s2+1, p2-s2);
		}
	}

@ Insert a character into the text buffer, ignoring any invalid characters.
Returns the number of characters inserted (0 or 1).

@proc int insert_char($, int c) {
	int i;
	Boolean okToPrint=False;

	if($valid == NULL) {		/* all printable chars are allowed */
		if(' ' <= c && c <= '~')
			okToPrint=True;
		}
	else {					/* check if c is a valid char */
		for(i=0;$valid[i];i++)
			if(c == $valid[i]) {
				okToPrint=True;
				break;
				}
		}

	if (!okToPrint) 
		return 0;

	/* remove any active text selection */
	if ($selected)
		if ($killSelection)
			delete_selection($);
		else
			clear_selection($);

	/* is there room to insert? */
	if($text_len == $max) {
		XBell(XtDisplay($),100);
		return 0;
		}

	/* now insert the char */
	$text_len++;
	for(i=$text_len;i > $text_pos; i--)
		$text[i]=$text[i-1];
	$text[$text_pos++]=c;

	return 1;
	}


@IMPORTS

@incl <assert.h>
@incl <ctype.h>
@incl <stdio.h>
@incl <string.h>
@incl <X11/Xatom.h>
@incl <X11/Xmu/Xmu.h>
