# Indicator widget
# Grant McDorman <grant@isgtec.com>
# Version 1.0

@class XfwfIndicator (XfwfLabel) @file=Indicator

@ A Indicator widget consists of a small slider sliding horizontally along
a scale. Part of the scale - the "range" - may have a different colour
background. A multi-line label is above the slider.

The position of the slider is not controlled by resources, 
but by function calls.

@public

@ The color of the slider is by default set to the default
background color of the display, but it can be changed with the |sliderColor|
resource.

	@var Pixel sliderColor = <String> XtDefaultBackground

@ The width of the slider is controlled by the |sliderWidth| resource.
(The height is automatically set according to the height of the widget).

	@var Dimension sliderWidth = 30

@ The color of the range is also the default background colour; it's controlled
by the resource |rangeColor|.

	@var Pixel rangeColor = <String> XtDefaultBackground

@ The routines on the callback list are called whenever the user
manipulates the indicator and also when the Indicator receives a call on
its |scrollResponse| function with a reason other than |XfwfSNotify|.

The |call_data| parameter of the callback routines is a pointer
to an |XfwfScrollInfo| structure, which looks like this: |typedef
struct _XfwfScrollInfo { XfwfSReason reason; XfwfSFlags flags; float
vpos, vsize, hpos, hsize;} XfwfScrollInfo|. Only |hpos| is meaningfull.

	@var <Callback> XtCallbackList scrollCallback = NULL

@ The Indicator widget provides has a method for dealing with scroll
requests from the application or from other widgets. A pointer to that
function can be retrieved with |XtGetValues| as the resource
|XtNscrollResponse|. This resource can only be queried, not set.

	@var XtCallbackProc scrollResponse = scroll_response

@ The width of the frame around the slider is independent of the frame
around the whole widget. It can be set with |sliderFrameWidth|.

	@var Dimension sliderFrameWidth = 2

@ The style of the frame around the slider is set with
|sliderFrameType|. By default, it is |XfwfRaised|. Note that there are no
resources to control the shadow scheme of the slider independently from
that of the outer frame. That means that the resources |shadowScheme|,
|topShadowColor|, |topShadowStipple|, etc, also influence the frame of the
slider.

	@var FrameType sliderFrameType = XfwfRaised

@ A hairline on the slider is drawn at its middle. Its width
is set with |hairlineWidth|.

	@var Dimension hairlineWidth = 1

@ The top and bottom margins of the hairline are  set with |harlineMargin|.
If this is zero, as the default, the hairline extends across the top and bottom
shadows of the slider. It cannot extend outside the slider.

	@var Dimension hairlineMargin = 0

@ The colour of the hair-line is set with |hairlineColor|.

	@var Pixel hairlineColor = XtDefaultForeground

@ The distance between the range area and the bottom of the widget is set
with |bottomMargin|.

	@var Dimension bottomMargin = 1

@ The distance between the range area and the label is set
with |labelMargin|.

	@var Dimension labelMargin = 1

@ The margin between the range area and the sliding slider is set with
|sliderMargin|.

	@var Dimension sliderMargin = 2

@ The default frame width of the indicator is changed from 0 to 2.

	@var frameWidth = 2

@ The default frame type is now |XfwfSunken|.

	@var frameType = XfwfSunken


@private

@ The position of the the slider is controlled by |slider_pos|.

@var float slider_pos

@var float rangeMin
@var float rangeMax

@ A boolean variable is set to when a dragging action has started, but
not yet finished.

@var Boolean drag_in_progress

@ During a drag operation, the slider is kept at a fixed offset from
the moving mouse. The offset is stored in a local variable.

@var int m_delta

@ We also need three more GC's for the slider and the light and dark
parts of the slider's frame, one for the range area, and one for the
hairline.

@var GC slidergc
@var GC sliderlightgc
@var GC sliderdarkgc
@var GC rangegc
@var GC hairlinegc

@exports

@ The |scroll.h| header file is needed for the |XfwfScrollInfo|
structure.

@incl <Xfwf/scroll.h>

@ The label can be set by |XfwfIndicatorSetLabel|. This is more efficient
than Xt[Va]SetValues. (Do not, however, change the number of lines in
the label).

@proc XfwfIndicatorSetLabel($, String newlabel)
{
    
    if (XtIsRealized($)) {
        Position    x, y;
        Dimension   w, h;
	Position    lx, ly;
	Dimension   lw, lh;
        
        #compute_inside($, &x, &y, &w, &h);
	$compute_inside($, &lx, &ly, &lw, &lh);
	if (lw != 0 && lh != 0) {
	    XClearArea(XtDisplay($), XtWindow($), x, y,
	               w, lh, FALSE);
	}
    }
    $set_label($, newlabel);
}

@ To change the position of the slider, a call to |XfwfIndicatorMovePoint|
can be used.  The arguments must be a number between 0 and 1.  The slider
is moved with |XCopyArea|.

This is a convenience function. The standard interface is through the
|scrollResponse| resource. In the Indicator widget, that resource is
connected to the |scroll_response| method.

@proc XfwfIndicatorMovePoint($, double pos)
{
    XfwfScrollInfo info;

    if (! XtIsSubclass($, xfwfIndicatorWidgetClass))
	XtError("XfwfMovePoint called with incorrect widget type");
    if (pos < 0.0 || pos > 1.0)
	XtError("XfwfMovePoint called with incorrect arguments");

    info.flags = XFWF_HPOS;
    info.reason = XfwfSNotify;
    info.hpos = pos;
    $scroll_response(NULL, $, &info);
}

@ To get the value of the point, use |XfwfIndicatorGetPoint|.

@proc XfwfIndicatorGetPoint($, double *pos)
{
    *pos = $slider_pos;
}

@ Setting the range is done with |XfwfIndicatorSetRange|. The arguments must be
between 0 and 1. If the minimum exceeds a maximum, the range is not
drawn.

@proc XfwfIndicatorSetRange($, double new_min, double new_max)
{
    Position old_x, new_x, y;
    Dimension old_width, new_width, height;
    Position rect_x, rect_y;
    Dimension rect_w, rect_h;
    Position fx, fy;
    Dimension fw, fh;
    Region	clip_region, sub_region;
    XRectangle	rect;

    if (XtIsRealized($))
	$compute_range($, &old_x, &y, &old_width, &height);

    if (new_min < 0.0 || new_min > 1.0)
    {
	XtAppWarning(XtWidgetToApplicationContext($),
		     "Indicator range minimum must be between 0.0 and 1.0");
	new_min = new_min < 0 ? 0.0 : 1.0;
    }
    if (new_max < 0.0 || new_max > 1.0)
    {
	XtAppWarning(XtWidgetToApplicationContext($),
		     "Indicator range maximum must be between 0.0 and 1.0");
	new_max = new_max < 0 ? 0.0 : 1.0;
    }
    $rangeMin = new_min;
    $rangeMax = new_max;

    if (!XtIsRealized($))
	return;

    $compute_range($, &new_x, &y, &new_width, &height);
    $compute_slider($, &rect_x, &rect_y, &rect_w, &rect_h,
			&fx, &fy, &fw, &fh);

    if (old_width != 0)
    {
	if (old_x < new_x) {
	    clear_area($, old_x, y, new_x - old_x, height,
		       rect_x, rect_y, rect_w, rect_h);
	}

	if (old_x + old_width > new_x + new_width) {
	    clear_area($, new_x + new_width, y,
		    old_x + old_width - (new_x + new_width),
		       height, rect_x, rect_y, rect_w, rect_h);
	}
    }

    /* Use a region to clip out the slider and do a FillArea. Dumb
	but simple (to code, at least...). */
    if (new_width == 0)
	return;


    clip_region = XCreateRegion();
    rect.x = new_x;
    rect.y = y;
    rect.width = new_width;
    rect.height = height;
    XUnionRectWithRegion(&rect, clip_region, clip_region);
    sub_region = XCreateRegion();
    rect.x = rect_x;
    rect.y = rect_y;
    rect.width = rect_w;
    rect.height = rect_h;
    XUnionRectWithRegion(&rect, sub_region, sub_region);
    XSubtractRegion(clip_region, sub_region, clip_region);
    XDestroyRegion(sub_region);

    XSetRegion(XtDisplay($), $rangegc, clip_region);
    XFillRectangle(XtDisplay($), XtWindow($), $rangegc,
		new_x, y, new_width, height);
    XSetClipMask(XtDisplay($), $rangegc, None);

    XDestroyRegion(clip_region);
}

@ The convienence function |XfwfIndicatorGetRange| returns the range limits.

@proc XfwfIndicatorGetRange($, double *ret_min, double *ret_max)
{
    *ret_min = $rangeMin;
    *ret_max = $rangeMax;
}

@methods

@ The |compute_slider| method returns the position and size of the
slider in pixels.

@proc compute_slider($, Position *x, Position *y,
		    Dimension *width, Dimension *height,
		    Position *fx, Position *fy,
		    Dimension *fw, Dimension *fh)
{

    #compute_inside($, fx, fy, fw, fh);

    *height = *fh - $label_height - $bottomMargin - $sliderMargin * 2 -
	$labelMargin;
    *width = $sliderWidth;
    if (*height > *fh || *height < 1)
	*height = *fh;
    if (*width > *fw)
	*width = *fw;
    *x = (int) (0.5 + *fx + (*fw - *width) * $slider_pos);
    *y = *fy + *fh - *height - $bottomMargin - $sliderMargin;
}

@ The |compute_range| method returns the position and size of the
range in pixels.

@proc compute_range($, Position *x, Position *y,
		    Dimension *width, Dimension *height)
{
    Position fx, fy;
    Dimension fw, fh;
    Position slider_x, slider_y;
    Dimension slider_w, slider_h;
    Position range_end;

    $compute_slider($,&slider_x, &slider_y, &slider_w, &slider_h,
		      &fx, &fy, &fw, &fh);
    *height = slider_h + 2 * $sliderMargin;
    *x = (int) (0.5 + fx + (fw - slider_w) * $rangeMin + 0.5 * slider_w);
    range_end = (int) (0.5 + fx + 
			(fw - slider_w) * $rangeMax + 0.5 * slider_w);
    *width = range_end - *x;
    if (range_end < *x)
	*width = 0;
    *y = slider_y - $sliderMargin;
}

@ The |compute_inside| method of the Label class returns the area inside the
frame, but the label of Indicator widget should appear above the slider.
Therefore the |compute_inside| method is redefined. This means that the |expose|
method of the Label class can still be used, because it calls this function
to establish the position of the text.

@proc compute_inside
{
    Position fx, fy;
    Dimension fw, fh;

    $compute_slider($, x, y, w, h, &fx, &fy, &fw, &fh);

    *y = fy;
    *x -= $label_width / 2 - *w / 2;
    *h = $label_height;
    *w = $label_width;
    if (*w > fw)
	*w = fw;
    if (*x < fx)
	*x = fx;
    if (*x + *w > fx + fw)
	*x = fx + fw - *w;
}

@ The |expose| method of the superclass is called to draw the outer frame
and the text above the slider.

@proc expose
{
    Position x, y;
    Position fx, fy;
    Position range_x, range_y;
    Dimension width, height;
    Dimension fw, fh;
    Dimension range_w, range_h;

    if (! XtIsRealized($)) return;
    if (region != NULL) {
	XSetRegion(XtDisplay($), $rangegc, region);
	XSetRegion(XtDisplay($), $slidergc, region);
	XSetRegion(XtDisplay($), $sliderdarkgc, region);
	XSetRegion(XtDisplay($), $sliderlightgc, region);
    }
    $compute_slider($, &x, &y, &width, &height, &fx, &fy, &fw, &fh);
    $compute_range($, &range_x, &range_y, &range_w, &range_h);
    if (range_w > 0)
	XFillRectangle(XtDisplay($), XtWindow($), $rangegc,
			range_x, range_y, range_w, range_h);

    XFillRectangle(XtDisplay($), XtWindow($), $slidergc,
		    x, y, width, height);
    XfwfDrawFrame($, x, y, width, height,
		$sliderFrameType, $sliderFrameWidth,
		$sliderlightgc, $sliderdarkgc);
    XFillRectangle(XtDisplay($), XtWindow($), $hairlinegc,
		    (x + width / 2 - $hairlineWidth / 2), y + $hairlineMargin,
		    $hairlineWidth, height - 2 * $hairlineMargin);

    if (region != NULL) {
	XSetClipMask(XtDisplay($), $rangegc, None);
	XSetClipMask(XtDisplay($), $slidergc, None);
	XSetClipMask(XtDisplay($), $sliderdarkgc, None);
	XSetClipMask(XtDisplay($), $sliderlightgc, None);
    }
    #expose($, event, region);
}

@ The |initialize| method only needs to set the local variables.

@proc initialize
{
    $slider_pos = 0.0;
    $rangeMin = 0.0;
    $rangeMax = 0.0;
    $drag_in_progress = False;
    $slidergc = NULL; create_slidergc($);
    $rangegc = NULL; create_rangegc($);
    $sliderlightgc = NULL; create_sliderlightgc($);
    $sliderdarkgc = NULL; create_sliderdarkgc($);
    $hairlinegc = NULL; create_hairlinegc($);
}

@ The following routine's name, |move_slider|, indicates what it is
used for, but not what it really does. It doesn't depend on the slider
at all, it simply copies a slider to another position in the window
and clears the old slider to the background color.

@proc move_slider($, int oldx, int y, int wd, int ht, int newx)
{
    int h;
    Position range_x, range_y;
    Dimension range_w, range_h;
    Position clear_x, fill_x;
    Dimension clear_w, fill_w;

    if (oldx == newx)
	return;		/* no change */

    $compute_range($, &range_x, &range_y, &range_w, &range_h);

    XCopyArea(XtDisplay($), XtWindow($), XtWindow($),
	      DefaultGCOfScreen(XtScreen($)),
	      oldx, y, wd, ht, newx, y);

    if (oldx < newx) {
	/* move right */
	if (oldx + wd > newx)
	    wd = newx - oldx;
    } else {
	if (newx + wd > oldx) {
	    Position tmp;

	    tmp = newx + wd;
	    wd = oldx - newx;
	    oldx = tmp;
	}
    }

    if (range_w == 0) {
	clear_x = oldx;
	clear_w = wd;
	fill_w = 0;
    } else {
	if (oldx < range_x) {
	    /* starts before range */
	    clear_x = oldx;
	    if (oldx + wd < range_x) {
		clear_w = wd;
		fill_w = 0;
	    } else {
		clear_w = range_x - oldx;
		fill_x = range_x;
		fill_w = wd - clear_w;
	    }
	} else {
	    /* starts in or after range */
	    if (oldx > range_x + range_w)
	    {
		clear_x = oldx;
		clear_w = wd;
		fill_w = 0;
	    } else {
		/* starts in range */
		fill_x = oldx;
		if (oldx + wd <= range_x + range_w) {
		    fill_w = wd;
		    clear_w = 0;
		} else {
		    fill_w = (range_x + range_w) - oldx;
		    clear_w = wd - fill_w;
		    clear_x = range_x + range_w;
		}
	    }
	}
    }
    if (clear_w > 0)
	XClearArea(XtDisplay($), XtWindow($), clear_x, y, clear_w, ht, False);
    if (fill_w > 0)
	XFillRectangle(XtDisplay($), XtWindow($), $rangegc, fill_x, y,
			fill_w, ht);
}


@ The |move_label| routine is similar to |move_slider|, but it moves the
label.

@proc move_label($, int oldx, int y, int wd, int ht, int newx)
{
    int h;

    if (oldx == newx)
	return;		/* no change */

    XCopyArea(XtDisplay($), XtWindow($), XtWindow($),
	      DefaultGCOfScreen(XtScreen($)),
	      oldx, y, wd, ht, newx, y);

    if (oldx < newx) {
	/* move right */
	if (oldx + wd > newx)
	    wd = newx - oldx;
    } else {
	if (newx + wd > oldx) {
	    Position tmp;

	    tmp = newx + wd;
	    wd = oldx - newx;
	    oldx = tmp;
	}
    }

    if (wd > 0)
	XClearArea(XtDisplay($), XtWindow($), oldx, y, wd, ht, False);
}


@ The |compute_info| method computes the relative position of the
slider, given its geometry in pixels. Before that, it makes sure the pixel
values are within the frame and it adapts the values if needed.

@proc compute_info($, Position *x, Position *y, Dimension *w, Dimension *h,
    float *slider_pos)
{
    Dimension fw, fh;
    Position fx, fy;

    #compute_inside($, &fx, &fy, &fw, &fh);

    if (*x < fx)
	*x = fx;
    if (*x + *w > fx + fw) 
	*x = fx + (fw - *w);

    *slider_pos = (*x - fx) / (float) (fw - *w);
}

@ The |set_values| method changes the GC's when needed.
A change in |minsize| doesn't necessarily cause a redraw; only if
the current slider size is less than the new minimum does the widget
needs to be redrawn.

@proc set_values
{
    Position x, y;
    Dimension w, h;
    Dimension fw, fh;
    Position fx, fy;

    if ($sliderColor != $old$sliderColor) {
	create_slidergc($);
    }
    if ($hairlineColor != $old$hairlineColor) {
	create_hairlinegc($);
    }
    if ($rangeColor != $old$rangeColor) {
	create_rangegc($);
    }
    if ($bottomShadowColor != $old$bottomShadowColor) {
	create_sliderdarkgc($);
    }
    if ($topShadowColor != $old$topShadowColor) {
	create_sliderlightgc($);
    }
    if ($scrollResponse != $old$scrollResponse) {
	$scrollResponse = $old$scrollResponse;
	XtWarning("scrollResponse resource may only be queried, not set");
    }
    return True;
}

@ The |clear_area| procedure clears an area ... except for the slider.
@proc clear_area($, Position x, Position y, Dimension width, Dimension height,
		 Position rect_x, Position rect_y,
		 Dimension rect_w, Dimension rect_h)
{
    Dimension clear_w;
    Position	start;

    if (width == 0)
	return;

    if (x < rect_x) {
	/* clear from start to slider or end, whichever first */

	clear_w = width;
	if (x + clear_w > rect_x)
	    clear_w = rect_x - x;
	if (clear_w != 0)
	    XClearArea(XtDisplay($), XtWindow($), x, y, clear_w, height, False);
    }
    if (x <= rect_x + rect_w && x + width >= rect_x) {
	/* clear strip above/below slider */

	clear_w = rect_w;
	start = rect_x;
	if (x >= rect_x && x + width < rect_x + rect_w) { /* in rect. */
	    clear_w = width;
	    start = x;
	} else if (x > rect_x && x + width >= rect_x + rect_w) { /* starts in */
	    clear_w = rect_x + rect_w - x;
	    start = x;
	} else if (x < rect_x && x + width < rect_x + rect_w) { /* ends in..*/
	    clear_w = x + width - rect_x;
	    start = rect_x;
	}
	if (rect_y > y && clear_w > 0)
	    XClearArea(XtDisplay($), XtWindow($), start, y,
			clear_w, rect_y - y, False);
	if (rect_y + rect_h < y + height && clear_w > 0)
	    XClearArea(XtDisplay($), XtWindow($), start, rect_y + rect_h,
			clear_w, y + height - (rect_y + rect_h), False);
    }
    if (x + width >= rect_x + rect_w) {
	/* clear from slider end or start, whichever last, to end */
	clear_w = width;
	start = x;
	if (x < rect_x + rect_w) {
	    start = rect_x + rect_w;
	    clear_w = x + width - (rect_x + rect_w);
	}
	if (clear_w > 0)
	    XClearArea(XtDisplay($), XtWindow($), start, y, clear_w,
			height, False);
    }
}

@ The method |scroll_response| is exported via the |scrollResponse|
resource. It has the format of a callback function, so that it can be
registered as a callback in the |scrollCallback| list of some other
widget. The |client_data| must be a pointer to the Indicator widget
itself, the |call_data| is a pointer to an |XfwfScrollInfo| structure.
The widget |wdg| is the widget from whose callback list the function
is called.

@def range(x) = (0.0 <= (x) && (x) <= 1.0)

@proc scroll_response(Widget wdg, XtPointer client_data, XtPointer call_data)
{
    Widget self = (Widget) client_data;
    XfwfScrollInfo *inf = (XfwfScrollInfo *)call_data;
    XfwfScrollInfo new_info;
    float x;
    Position newx, y, oldx;
    Dimension newwd, newht, oldwd, oldht, wd, ht;
    XEvent event;
    XRectangle rect;
    Region clip;
    Display *dpy = XtDisplay($);
    Dimension fw, fh;
    Position fx, fy;
    Position lbl_oldx, lbl_y, lbl_newx;
    Dimension lbl_wd, lbl_ht;

    x = (inf->flags&XFWF_HPOS) && range(inf->hpos) ? inf->hpos : $slider_pos;
    if ($slider_pos != x) { /* Position changed */
	if (XtIsRealized($))
	{
	    $compute_slider($, &oldx, &y, &wd, &ht, &fx, &fy, &fw, &fh);
	    $compute_inside($, &lbl_oldx, &lbl_y, &lbl_wd, &lbl_ht);
	}
	$slider_pos = x;
	if (XtIsRealized($)) {
	    $compute_slider($, &newx, &y, &wd, &ht, &fx, &fy, &fw, &fh);
	    $compute_inside($, &lbl_newx, &lbl_y, &lbl_wd, &lbl_ht);
	    $move_slider($, oldx, y, wd, ht, newx);
	    $move_label($, lbl_oldx, lbl_y, lbl_wd, lbl_ht, lbl_newx);
	}
    }

    if (inf->reason != XfwfSNotify) {
	new_info = *inf;
	new_info.reason = XfwfSNotify;
	XtCallCallbackList($, $scrollCallback, &new_info);
    }
}

@translations

@ The |start| action should be bound to a mouse button press, because it
needs the coordinates of the mouse. The |drag| action is bound to mouse
movement and the |finish| action is normally bound to a release  of the
mouse button.

@trans <Btn1Down>: start()
@trans <Btn1Motion>: drag()
@trans <Btn1Up>: finish()


@actions

@ The |start| action checks the position of the mouse and if it was
outside the slider, it calls the |scrollCallback|. Otherwise, it only
records the position. Note that the mouse may have been to the left as
well as below the slider, causing the callbacks to be called twice.

@proc start
{
    Dimension w, h;
    Position x, y;
    XfwfScrollInfo info;
    Boolean outside = False;
    Dimension fw, fh;
    Position fx, fy;


    if (event->type != ButtonPress && event->type != ButtonRelease
	&& event->type != MotionNotify)
	XtError("The start action must be bound to a mouse event");
    $compute_slider($, &x, &y, &w, &h, &fx, &fy, &fw, &fh);
    if (event->xbutton.x < x) {			/* Left of slider */
	info.reason = XfwfSPageLeft;
	outside = True;
    }
    if (event->xbutton.x >= x + w) {		/* Right of slider */
	info.reason = XfwfSPageRight;
	outside = True;
    }
    if (outside) {
	Dimension wd, ht, fw, fh;
	Position oldx, y, newx, fx, fy;
	Position lbl_oldx, lbl_y, lbl_newx;
	Dimension lbl_wd, lbl_ht;

	$compute_slider($, &oldx, &y, &wd, &ht, &fx, &fy, &fw, &fh);
	$compute_inside($, &lbl_oldx, &lbl_y, &lbl_wd, &lbl_ht);

	newx = event->xbutton.x;

	$compute_info($, &newx, &y, &wd, &ht, &$slider_pos);
	$compute_inside($, &lbl_newx, &lbl_y, &lbl_wd, &lbl_ht);
	$move_slider($, oldx, y, wd, ht, newx);
	$move_label($, lbl_oldx, lbl_y, lbl_wd, lbl_ht, lbl_newx);

	info.hpos = $slider_pos;

    	XtCallCallbackList($, $scrollCallback, &info);

	$m_delta = 0;
    } else
	$m_delta = x - event->xbutton.x;

    $drag_in_progress = True;
}

@ The |finish| action does nothing if this is the end of a
click outside the slider. The callbacks for this event have already
been called.

If this is the end of a drag action, we reset the flag
|drag_in_progress| to False and call the drop callbacks.

@proc finish
{
    XfwfScrollInfo info;

    if ($drag_in_progress) {
	$drag_in_progress = False;
	info.reason = XfwfSMove;
	info.flags = XFWF_HPOS;
	info.hpos = $slider_pos;
	XtCallCallbackList($, $scrollCallback, &info);
    }
}

@ An application that can draw fast enough, may wish to redraw with
every movement of the slider, instead of only at the end of the drag
action. The drag callback is provided for this purpose. It is called
in the same way as the drop callback, with the current relative
position of the slider.

@proc drag
{
    XfwfScrollInfo info;
    Dimension wd, ht, fw, fh;
    Position oldx, y, newx, fx, fy;
    Position lbl_oldx, lbl_y, lbl_newx;
    Dimension lbl_wd, lbl_ht;

    if (! $drag_in_progress) return;
    if (event->type != ButtonPress && event->type != ButtonRelease
	&& event->type != MotionNotify)
	XtError("The drag action must be bound to a mouse event");

    $compute_slider($, &oldx, &y, &wd, &ht, &fx, &fy, &fw, &fh);
    $compute_inside($, &lbl_oldx, &lbl_y, &lbl_wd, &lbl_ht);

    newx = event->xbutton.x + $m_delta;

    $compute_info($, &newx, &y, &wd, &ht, &$slider_pos);
    $compute_inside($, &lbl_newx, &lbl_y, &lbl_wd, &lbl_ht);
    $move_slider($, oldx, y, wd, ht, newx);
    $move_label($, lbl_oldx, lbl_y, lbl_wd, lbl_ht, lbl_newx);

    info.reason = XfwfSDrag;
    info.flags = XFWF_HPOS;
    info.hpos = $slider_pos;
    XtCallCallbackList($, $scrollCallback, &info);
}


@utilities


@ |create_slidergc| creates the GC that draw the background in the
slider.

@proc create_slidergc($)
{
    XtGCMask mask;
    XGCValues values;

    if ($slidergc != NULL) XtReleaseGC($, $slidergc);
    mask = GCForeground;
    values.foreground = $sliderColor;
    $slidergc = XtGetGC($, mask, &values);
}

@ The |create_sliderlightgc| functions makes the GC for drawing the light
parts of the slider's frame.

@proc create_sliderlightgc($)
{
    XtGCMask mask;
    XGCValues values;

    if ($sliderlightgc != NULL) XtReleaseGC($, $sliderlightgc);
    switch ($shadowScheme) {
    case XfwfColored:
	mask = GCForeground;
	values.foreground = $topShadowColor;
	break;
    case XfwfStipple:
	mask = GCFillStyle | GCStipple | GCForeground | GCBackground;
	values.fill_style = FillOpaqueStippled;
	values.background = $sliderColor;
	values.stipple = $topShadowStipple;
	values.foreground = WhitePixelOfScreen(XtScreen($));
	break;
    case XfwfAuto:
	if (DefaultDepthOfScreen(XtScreen($)) > 4
	    && $lighter_color($, $sliderColor, &values.foreground)) {
	    mask = GCForeground;
	} else {
	    mask = GCFillStyle | GCBackground | GCForeground | GCStipple;
	    values.fill_style = FillOpaqueStippled;
	    values.background = $sliderColor;
	    values.foreground = WhitePixelOfScreen(XtScreen($));
	    values.stipple =
		XCreateBitmapFromData(XtDisplay($),
				      RootWindowOfScreen(XtScreen($)),
				      stip4_bits, stip4_width, stip4_height);
	}
	break;
    }
    $sliderlightgc = XtGetGC($, mask, &values);
}

@ The |create_sliderdarkgc| routines does the same for the dark parts of the
slider's frame.

@proc create_sliderdarkgc($)
{
    XtGCMask mask;
    XGCValues values;

    if ($sliderdarkgc != NULL) XtReleaseGC($, $sliderdarkgc);
    switch ($shadowScheme) {
    case XfwfColored:
	mask = GCForeground;
	values.foreground = $bottomShadowColor;
	break;
    case XfwfStipple:
	mask = GCFillStyle | GCStipple | GCForeground | GCBackground;
	values.fill_style = FillOpaqueStippled;
	values.stipple = $bottomShadowStipple;
	values.foreground = BlackPixelOfScreen(XtScreen($));
	values.background = $sliderColor;
	break;
    case XfwfAuto:
	if (DefaultDepthOfScreen(XtScreen($)) > 4
	    && $darker_color($, $sliderColor, &values.foreground)) {
	    mask = GCForeground;
	} else {
	    mask = GCFillStyle | GCBackground | GCForeground | GCStipple;
	    values.fill_style = FillOpaqueStippled;
	    values.background = $sliderColor;
	    values.foreground = WhitePixelOfScreen(XtScreen($));
	    values.stipple =
		XCreateBitmapFromData(XtDisplay($),
				      RootWindowOfScreen(XtScreen($)),
				      stip4_bits, stip4_width, stip4_height);
	}
	break;
    }
    $sliderdarkgc = XtGetGC($, mask, &values);
}

@ |create_rangegc| creates the GC that draw the background in the
range.

@proc create_rangegc($)
{
    XtGCMask mask;
    XGCValues values;

    if ($rangegc != NULL) XtReleaseGC($, $rangegc);
    mask = GCForeground;
    values.foreground = $rangeColor;
    $rangegc = XtGetGC($, mask, &values);
}

@ |create_hairlinegc| creates the GC that draws the hairline.

@proc create_hairlinegc($)
{
    XtGCMask mask;
    XGCValues values;

    if ($hairlinegc != NULL) XtReleaseGC($, $hairlinegc);
    mask = GCForeground;
    values.foreground = $hairlineColor;
    $hairlinegc = XtGetGC($, mask, &values);
}

@imports

@incl "stip4.bm"
@incl <stdio.h>
