# PcBar - progress bar widget.
# Author: Lachlan Wetherall <lwether@cs.newcastle.edu.au>
#
#    .-_|\   Department of Computer Science
#   /     \  University of Newcastle
#   \.--._*  Callaghan NSW 2308
#        v   Australia
#
# Revision History:
# Date       Ver    Comment
# ----       ---    -------
# 16-Dec-94  1.0    Original release
# 06-Sep-96  1.1    Rebuilt using wbuild and now a subclass of Board.
#                   Added textcolor and usetextcolor resources.


@class XfwfPcBar (XfwfBoard)     @file=PcBar


@public

@ The foreground color.

	@var Pixel foreground = <String> XtDefaultForeground

@ The color in which to draw the text if |displaypc| and |usetextcolor| are
both true.

	@var Pixel textcolor = <String> "blue"

@ The percentage value to be displayed.

	@var int percentage = 0

@ Whether to display a percentage in text.

	@var Boolean displaypc = False

@ If displaying percentage as text, whether to display a 0 value.

	@var Boolean showzero = False

@ If true, the bar will be vertical rather than horizontal.

	@var Boolean vertical = False

@ If |displaypc| is true, then if |usetextcolor| is true then the text will be
colored acording to the |textcolor| resource, otherwise it will be
colored using the |foreground| and |background| colors as appropriate.

	@var Boolean usetextcolor = False

@ The font to display the text in.

	@var <FontStruct> XFontStruct *font = <String> XtDefaultFont

@ The border width defaults to 1.

	@var border_width = 1

@PRIVATE

@ The GC for drawing.

        @var GC draw_gc

@ The GC for writing.

        @var GC text_gc

@utilities

@ Initialize the GC for drawing.

@proc GetGC($) {
    XGCValues values;
    XtGCMask mask = GCForeground | GCBackground;

    values.foreground = $foreground;
    values.background = $background_pixel;

    $draw_gc = XtGetGC($, mask, &values);

	if ($usetextcolor) {
		values.foreground = $textcolor;
		mask = GCForeground | GCBackground;
 		}
	else {
		values.function = GXxor;
		values.foreground = $foreground ^ $background_pixel;
		mask = GCFunction | GCForeground | GCBackground;
		}
    $text_gc = XtGetGC($, mask, &values);
	XSetFont(XtDisplay($), $text_gc, $font->fid); 
	}

@METHODS

@def H_DEFAULTWIDTH = 200  
@def H_DEFAULTHEIGHT = 20  
@def V_DEFAULTWIDTH = 20
@def V_DEFAULTHEIGHT = 200  

@proc initialize {

    if (($percentage < 0) ||	($percentage > 100))  {
		XtWarning("PcBar: invalid percentage (using 0)."); 
    	$percentage = 0;
	    }

    GetGC($);
    if ($width == 0) 
		$width = $vertical ? V_DEFAULTWIDTH : H_DEFAULTWIDTH;
    if ($height == 0) 
		$height = $vertical ? V_DEFAULTHEIGHT : H_DEFAULTHEIGHT;

	}

@proc expose {

    if (!XtIsRealized($))
		return;

	XfwfPcBarSetPercentage($, $percentage);
	}


@proc set_values {
    Boolean do_redisplay = False;

    if (($old$foreground != $foreground) ||
    	($old$background_pixel != $background_pixel) ||
    	($old$textcolor != $textcolor) ||
    	($old$usetextcolor != $usetextcolor) ||
     	($old$font != $font)) {
		XtReleaseGC($, $draw_gc);
		XtReleaseGC($, $text_gc);
		GetGC($);
		do_redisplay = True;
	    }
    if ($old$percentage != $percentage) 
		do_redisplay = True;
    if ($old$displaypc != $displaypc) 
		do_redisplay = True;
    if (($old$showzero != $showzero) && ($percentage == 0)) {
		do_redisplay = True;
		}
    if ($old$vertical != $vertical) 
		do_redisplay = True;

    return do_redisplay;
	}


@proc destroy {
    if ($draw_gc)
		XtReleaseGC($, $draw_gc);
    if ($text_gc)
		XtReleaseGC($, $text_gc);
	}



@proc query_geometry {
	/* accept any requested change in size */
    return XtGeometryYes;
	}

@EXPORTS

@ The |XfwfPcBarSetPercentage| convenience function sets the percentage
level to be displayed. Using this function rather than a call to
|XtSetValues| will result in less flicker if the percentage value is
changed quickly.

@proc void XfwfPcBarSetPercentage($, int pc) {
    Display *disp;
    Window win;
	int x,y;
	int tx,ty;
	char buf[10];
	int asc, des, dir;
	XCharStruct overall;

	$percentage=pc;
    if (!XtIsRealized($))
		return;

	disp=XtDisplay($);
	win=XtWindow($);

    XSetLineAttributes(disp, $draw_gc, $height, LineSolid,CapButt,JoinBevel);
	if ($vertical) {
	    XSetLineAttributes(disp, $draw_gc, $width, LineSolid,CapButt,JoinBevel);
	    y= $height - ($percentage * $height)/100;
	    XDrawLine(disp, win,$draw_gc,  $width/2, y, $width/2, $height);
    	if ($percentage <100) 
			XClearArea(disp, win, 0, 0, $width, y, False);
		}
	else {
	    XSetLineAttributes(disp,$draw_gc, $height, LineSolid,CapButt,JoinBevel);
	    x=($percentage * $width)/100;
	    XDrawLine(disp, win, $draw_gc, 0, $height/2, x, $height/2);
    	XClearArea(disp, win, x, 0, $width, $height, False);
		}

	if (($displaypc) && ($showzero || $percentage > 0)) {
		sprintf(buf, "%d%%", $percentage);
        XTextExtents($font, buf, strlen(buf), &dir, &asc, &des, &overall);
		tx = ($width-overall.width)/2;
		ty = ($height - (asc+des))/2 + asc;
		XDrawString(disp, win, $text_gc, tx, ty,  buf, strlen(buf));
		}
	}

@imports

@incl <stdio.h>
